/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *   * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */

package org.wso2.registry.secure;

import org.wso2.registry.ActionConstants;
import org.wso2.registry.RegistryConstants;
import org.wso2.registry.RegistryException;
import org.wso2.registry.i18n.Messages;
import org.wso2.registry.jdbc.DatabaseConstants;
import org.wso2.registry.jdbc.ConnectionFactory;
import org.wso2.usermanager.*;
import org.wso2.usermanager.UserManagerConstants;
import org.wso2.usermanager.readwrite.DefaultRealm;
import org.wso2.usermanager.readwrite.DefaultRealmConfig;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import javax.sql.DataSource;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

public class RegistryRealm {

    private static final Log log = LogFactory.getLog(RegistryRealm.class);

    public static Realm createRegistryRealm() throws RegistryException {

        Realm realm;

        DataSource umDataSource = getUserManagerDataSource();
        if (umDataSource != null) {
            realm = createDataSourceAwareRegistryRealm(umDataSource);
        } else {
            realm = createInMemoryRegistryRealm();
        }

        return realm;
    }

    public static Realm createInMemoryRegistryRealm() throws RegistryException {
        new HSQLDBInitializer().createHSQLTables();
        try {
            DefaultRealm realm = new DefaultRealm();
            DefaultRealmConfig config = (DefaultRealmConfig)realm.getRealmConfiguration();
            config.setConnectionURL(RegistryUMConstants.HSQL_DB_URL);
            config.setDriverName(DatabaseConstants.HSQL_DRIVER_NAME);
            realm.init(config);
            // define built in roles
            //populateRoles(realm);
            return realm;

        } catch (Exception e) {
            throw new RegistryException(e.getMessage());
        }
    }

    public static Realm createDataSourceAwareRegistryRealm(DataSource umDataSource)
            throws RegistryException {

         try {
             DefaultRealm realm = new DefaultRealm();
             DefaultRealmConfig config = (DefaultRealmConfig)realm.getRealmConfiguration();
             config.setDataSource(umDataSource);
             realm.init(config);
             // define built in roles
            //populateRoles(realm);
            return realm;

        } catch (Exception e) {
            e.printStackTrace();
            throw new RegistryException(e.getMessage());
        }
    }

    private static void populateRoles(DefaultRealm realm) throws UserManagerException {

        UserStoreReader userStoreReader = realm.getUserStoreReader();
        UserStoreAdmin userStoreAdmin = realm.getUserStoreAdmin();
        AccessControlAdmin accessControlAdmin = realm.getAccessControlAdmin();
        
        String[] roles = userStoreReader.getAllRoleNames();

        if (!containsString(RegistryConstants.ADMIN_ROLE, roles)) {

            userStoreAdmin.addRole(RegistryConstants.ADMIN_ROLE);

            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, RegistryConstants.ROOT_PATH, ActionConstants.GET);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, RegistryConstants.ROOT_PATH, ActionConstants.PUT);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, RegistryConstants.ROOT_PATH, ActionConstants.DELETE);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, RegistryConstants.ROOT_PATH, ActionConstants.AUTHORIZE);

            // authorizations for user management
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, UserManagerConstants.USER_RESOURCE, UserManagerConstants.ADD);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, UserManagerConstants.USER_RESOURCE, UserManagerConstants.READ);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, UserManagerConstants.USER_RESOURCE, UserManagerConstants.EDIT);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, UserManagerConstants.USER_RESOURCE, UserManagerConstants.DELETE);

            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, UserManagerConstants.ROLE_RESOURCE, UserManagerConstants.ADD);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, UserManagerConstants.ROLE_RESOURCE, UserManagerConstants.READ);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, UserManagerConstants.ROLE_RESOURCE, UserManagerConstants.EDIT);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, UserManagerConstants.ROLE_RESOURCE, UserManagerConstants.DELETE);

            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, UserManagerConstants.PERMISSION_RESOURCE, UserManagerConstants.READ);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, UserManagerConstants.PERMISSION_RESOURCE, UserManagerConstants.ADD);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, UserManagerConstants.PERMISSION_RESOURCE, UserManagerConstants.EDIT);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, UserManagerConstants.PERMISSION_RESOURCE, UserManagerConstants.DELETE);

            //accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, RegistryConstants.ROOT_PATH, ActionConstants.ADD_USER);
            //accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, RegistryConstants.ROOT_PATH, ActionConstants.REMOVE_USER);
            //accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, RegistryConstants.ROOT_PATH, ActionConstants.ADD_ROLE);
            //accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE, RegistryConstants.ROOT_PATH, ActionConstants.REMOVE_ROLE);
        }

        if (!containsString(RegistryConstants.GUESTS_ROLE, roles)) {
            userStoreAdmin.addRole(RegistryConstants.GUESTS_ROLE);
        }

        if (!containsString(RegistryConstants.EVERYONE_ROLE, roles)) {
            userStoreAdmin.addRole(RegistryConstants.EVERYONE_ROLE);

            accessControlAdmin.authorizeRole(RegistryConstants.EVERYONE_ROLE, UserManagerConstants.USER_RESOURCE, UserManagerConstants.READ);
            accessControlAdmin.authorizeRole(RegistryConstants.EVERYONE_ROLE, UserManagerConstants.ROLE_RESOURCE, UserManagerConstants.READ);
            accessControlAdmin.authorizeRole(RegistryConstants.EVERYONE_ROLE, UserManagerConstants.PERMISSION_RESOURCE, UserManagerConstants.READ);            

            accessControlAdmin.authorizeRole(RegistryConstants.EVERYONE_ROLE, RegistryConstants.ROOT_PATH, ActionConstants.GET);
        }

        // define built in users

        String[] users = userStoreReader.getAllUserNames();

        if (!containsString(RegistryConstants.ADMIN_USER, users)) {

            userStoreAdmin.addUser(RegistryConstants.ADMIN_USER, "admin");

            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, RegistryConstants.ROOT_PATH, ActionConstants.GET);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, RegistryConstants.ROOT_PATH, ActionConstants.PUT);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, RegistryConstants.ROOT_PATH, ActionConstants.DELETE);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, RegistryConstants.ROOT_PATH, ActionConstants.AUTHORIZE);

            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, UserManagerConstants.USER_RESOURCE, UserManagerConstants.ADD);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, UserManagerConstants.USER_RESOURCE, UserManagerConstants.READ);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, UserManagerConstants.USER_RESOURCE, UserManagerConstants.EDIT);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, UserManagerConstants.USER_RESOURCE, UserManagerConstants.DELETE);

            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, UserManagerConstants.ROLE_RESOURCE, UserManagerConstants.ADD);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, UserManagerConstants.ROLE_RESOURCE, UserManagerConstants.READ);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, UserManagerConstants.ROLE_RESOURCE, UserManagerConstants.EDIT);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, UserManagerConstants.ROLE_RESOURCE, UserManagerConstants.DELETE);

            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, UserManagerConstants.PERMISSION_RESOURCE, UserManagerConstants.READ);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, UserManagerConstants.PERMISSION_RESOURCE, UserManagerConstants.ADD);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, UserManagerConstants.PERMISSION_RESOURCE, UserManagerConstants.EDIT);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, UserManagerConstants.PERMISSION_RESOURCE, UserManagerConstants.DELETE);

            // authorizations for registry level actions are set in the root path
            //accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, RegistryConstants.ROOT_PATH, ActionConstants.ADD_USER);
            //accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER, RegistryConstants.ROOT_PATH, ActionConstants.REMOVE_USER);
            //accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_ROLE, RegistryConstants.ROOT_PATH, ActionConstants.ADD_ROLE);
            //accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_ROLE, RegistryConstants.ROOT_PATH, ActionConstants.REMOVE_ROLE);
        }

        if (!containsString(RegistryConstants.ANONYMOUS_USER, users)) {
            userStoreAdmin.addUser(RegistryConstants.ANONYMOUS_USER, "guest");
            userStoreAdmin.addUserToRole(
                    RegistryConstants.ANONYMOUS_USER, RegistryConstants.EVERYONE_ROLE);
        }
    }

    public static DataSource getUserManagerDataSource() throws RegistryException {

        try {
            log.info("Trying to find the user manager data source " + DatabaseConstants.USER_MANAGER_DATASOURCE_NAME + "...");

            Context context = new InitialContext();
            DataSource  dataSource = (DataSource) context.lookup(DatabaseConstants.USER_MANAGER_DATASOURCE_NAME);

            log.info("Found the user manager data source " + DatabaseConstants.USER_MANAGER_DATASOURCE_NAME + ".");
            return dataSource;

        } catch (NamingException e) {
            log.info(Messages.
                    getMessage("datasource.not.found", DatabaseConstants.USER_MANAGER_DATASOURCE_NAME));
        }
        
        return null;
    }

    private static boolean containsString(String value, String[] array) {
        for (int i = 0; i < array.length; i++) {
            if (value.equals(array[i])) {
                return true;
            }
        }
        return false;
    }
}
